import 'package:clean_android_design/clean_android_design.dart';

class PathTransformerCollection {
  final PageCollectionMap pageCollectionMap;

  List<PathTransformer> _transformersPre = <PathTransformer>[];
  late List<PathTransformer> _transformersDefault;
  List<PathTransformer> _transformersPost = <PathTransformer>[];

  PathTransformerCollection({required this.pageCollectionMap})
      : _transformersDefault = [
          LeftPagePathTransformer(),
          DoublePagePathTransformer(),
          DialogBelowPathTransformer(),
          HomePathTransformer(pageCollectionMap),
        ];

  bool addPreDefault(PathTransformer transformer) {
    if (_transformersPre.any((x) => x.isEqual(transformer))) {
      return false;
    } else {
      _transformersPre.add(transformer);
      return true;
    }
  }

  bool addPostDefault(PathTransformer transformer) {
    if (_transformersPost.any((x) => x.isEqual(transformer))) {
      return false;
    } else {
      _transformersPost.add(transformer);
      return true;
    }
  }

  SplitRoutePath transformAll(SplitRoutePath path) {
    SplitRoutePath newPath = path.copy();

    for (var t in _transformersPre) newPath = t.transformPath(newPath);
    for (var t in _transformersDefault) newPath = t.transformPath(newPath);
    for (var t in _transformersPost) newPath = t.transformPath(newPath);

    return newPath;
  }
}

abstract class PathTransformer {
  SplitRoutePath transformPath(SplitRoutePath path);
  bool isEqual(PathTransformer pathTransformer);
}

class DoublePagePathTransformer implements PathTransformer {
  SplitRoutePath transformPath(SplitRoutePath path) {
    var list = <PageInstance>[];
    for (var instance in path.pageInstances.reversed) {
      if (instance is KeyPageInstance) {
        if (!list.any((x) =>
            (x is KeyPageInstance) && x.keyPage.key == instance.keyPage.key)) {
          list.add(instance);
        }
      } else {
        list.add(instance);
      }
    }

    return SplitRoutePath(list.reversed.toList());
  }

  bool isEqual(PathTransformer pathTransformer) =>
      pathTransformer is DoublePagePathTransformer;
}

class LeftPagePathTransformer implements PathTransformer {
  SplitRoutePath transformPath(SplitRoutePath path) {
    var list = <PageInstance>[];
    var gotOneLeft = false;
    for (var instance in path.pageInstances.reversed) {
      if (instance.position == SplitRouterPosition.Right) {
        if (!gotOneLeft) {
          list.add(instance);
        }
      } else {
        if (instance.position == SplitRouterPosition.Left) gotOneLeft = true;
        list.add(instance);
      }
    }

    return SplitRoutePath(list.reversed.toList());
  }

  bool isEqual(PathTransformer pathTransformer) =>
      pathTransformer is LeftPagePathTransformer;
}

class HomePathTransformer implements PathTransformer {
  final PageCollectionMap pageCollectionMap;

  HomePathTransformer(this.pageCollectionMap);

  SplitRoutePath transformPath(SplitRoutePath path) {
    if (path.pageInstances.length == 0)
      return SplitRoutePath([pageCollectionMap.pageHome.withoutParam()]);

    if (path.pageInstances.first.position != SplitRouterPosition.Left) {
      return SplitRoutePath([pageCollectionMap.pageHome.withoutParam()]
        ..addAll(path.pageInstances));
    }

    return path.copy();
  }

  bool isEqual(PathTransformer pathTransformer) =>
      pathTransformer is HomePathTransformer;
}

class DialogBelowPathTransformer implements PathTransformer {
  SplitRoutePath transformPath(SplitRoutePath path) {
    var list = <PageInstance>[];
    var gotOneNoneDialog = false;
    for (var instance in path.pageInstances.reversed) {
      if (instance.position == SplitRouterPosition.Dialog) {
        if (!gotOneNoneDialog) {
          list.add(instance);
        }
      } else {
        gotOneNoneDialog = true;
        list.add(instance);
      }
    }

    return SplitRoutePath(list.reversed.toList());
  }

  bool isEqual(PathTransformer pathTransformer) =>
      pathTransformer is LeftPagePathTransformer;
}
