import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class RequirementBuilder extends StatefulWidget {
  final SplitRoutePath routePath;
  final Widget Function(BuildContext context, SplitRoutePath routePath) builder;
  final Widget Function(BuildContext context) loadingBuilder;
  final void Function(SplitRoutePath invalidPath, PageInstance newPageInstance)
      redirectOnFail;

  const RequirementBuilder({
    Key? key,
    required this.routePath,
    required this.builder,
    required this.loadingBuilder,
    required this.redirectOnFail,
  }) : super(key: key);

  @override
  State<RequirementBuilder> createState() => RequirementBuilderState();
}

class RequirementBuilderState extends State<RequirementBuilder> {
  SplitRoutePath? lastCheckedPath;
  RequirementsResult? lastResult;

  @override
  Widget build(BuildContext context) {
    if (lastResult == null ||
        !(lastCheckedPath?.equals(widget.routePath) ?? false)) {
      lastResult = null;
      lastCheckedPath = widget.routePath.copy();

      var requirements = getRequirements();
      var resultDirect = checkDirect(requirements);
      if (resultDirect.result == RequirementsResultEnum.NeedAsync) {
        var copyOfPath = lastCheckedPath!.copy();
        checkAsync(requirements).then((resultAsync) {
          if (lastCheckedPath != null &&
              copyOfPath.equals(lastCheckedPath!) &&
              mounted)
            setState(() {
              lastResult = resultAsync;
            });
        });
      } else {
        lastResult = resultDirect;
      }
    }

    if (lastResult == null) return widget.loadingBuilder(context);

    if (lastResult!.result == RequirementsResultEnum.Success)
      return widget.builder(context, widget.routePath);

    if (lastResult!.result == RequirementsResultEnum.Fail) {
      if (lastResult!.useAsOverlay!) {
        return widget.builder(context, SplitRoutePath([lastResult!.failPage!]));
      } else {
        Future.delayed(
            Duration.zero,
            () =>
                widget.redirectOnFail(widget.routePath, lastResult!.failPage!));
      }
    }

    return widget.loadingBuilder(context);
  }

  void recheckRequirements() {
    print("[RequirementBuilder] Recheck requirements");
    if (mounted)
      setState(() {
        lastCheckedPath = null;
        lastResult = null;
      });
  }

  List<PageRequirement> getRequirements() {
    var requirements = <PageRequirement>[];
    widget.routePath.pageInstances
        .where((p) => p is KeyPageInstance)
        .cast<KeyPageInstance>()
        .map((p) => p.keyPage.requirements)
        .forEach((rs) {
      rs?.forEach((r) {
        if (!requirements.any((r2) => r2.isEqual(r))) {
          requirements.add(r);
        }
      });
    });
    return requirements;
  }

  RequirementsResult checkDirect(List<PageRequirement> requirements) {
    for (var requirement in requirements) {
      var result = requirement.fulfillRequirements();

      if (result.result == RequirementsResultEnum.Fail ||
          result.result == RequirementsResultEnum.NeedAsync) {
        return result;
      }
    }
    return RequirementsResult.success();
  }

  Future<RequirementsResult> checkAsync(
      List<PageRequirement> requirements) async {
    for (var requirement in requirements) {
      var result = await requirement.fulfillRequirementsAsync();
      if (result.result == RequirementsResultEnum.Fail) {
        return result;
      }
    }
    return RequirementsResult.success();
  }
}
