import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class SplitRouteInformationParser
    extends RouteInformationParser<SplitRoutePath> {
  static late SplitRouteInformationParser instance;

  SplitRouteInformationParser.newInstance({
    required this.pageCollectionMap,
  }) {
    instance = this;
  }

  final PageCollectionMap pageCollectionMap;

  SplitRouteInformationParser(this.pageCollectionMap);

  @override
  Future<SplitRoutePath> parseRouteInformation(
      RouteInformation routeInformation) async {
    var location = routeInformation.location ?? "/";
    location = location.endsWith("/")
        ? location.substring(0, location.length - 1)
        : location;
    final uri = Uri.parse(location);

    return parseInternalUri(uri);
  }

  SplitRoutePath parseInternalUri(Uri uri) {
    print("[SplitRouteInformationParser] parseInternalUri: ${uri.path}");
    List<PageInstance> instances = <PageInstance>[];

    if (uri.pathSegments.length == 0) {
      instances.add(pageCollectionMap.pageHome.withoutParam());
    } else {
      var segmentsLeft = <String>[]..addAll(uri.pathSegments);

      while (segmentsLeft.length != 0) {
        var key = segmentsLeft[0];
        var isAnonymousPage = key.length == 5 &&
            (int.tryParse(key)?.toString() ?? "x").padLeft(5, "0") == key;

        if (isAnonymousPage) {
          var currentVersion = SplitRouterDelegate
              .instance.routePath.pageInstances
              .where((p) => p is AnonymousPage)
              .cast<AnonymousPage?>()
              .firstWhere((p) => p!.keyNumber == key, orElse: () => null);
          if (currentVersion != null) {
            instances.add(currentVersion);
          }
          segmentsLeft = segmentsLeft.sublist(1);
        } else {
          if (pageCollectionMap.containsKey(segmentsLeft[0])) {
            var keyPage = pageCollectionMap[key];
            segmentsLeft = segmentsLeft.sublist(1);
            if (segmentsLeft.length >= keyPage.parameterCount) {
              var params = segmentsLeft.sublist(0, keyPage.parameterCount);

              var currentVersion = SplitRouterDelegate
                  .instance.routePath.pageInstances
                  .where((p) => p is KeyPageInstance)
                  .cast<KeyPageInstance?>()
                  .firstWhere(
                      (p) =>
                          p!.keyPage.key == key &&
                          p.param.join("/") == params.join("/"),
                      orElse: () => null);

              if (currentVersion != null) {
                instances.add(currentVersion);
              } else {
                instances.add(keyPage.withParam(params));
              }

              segmentsLeft = segmentsLeft.sublist(keyPage.parameterCount);
            } else {
              return SplitRoutePath([pageCollectionMap.page404.withoutParam()]);
            }
          } else {
            return SplitRoutePath([pageCollectionMap.page404.withoutParam()]);
          }
        }
      }
    }

    return SplitRoutePath(instances);
  }

  @override
  RouteInformation restoreRouteInformation(SplitRoutePath path) {
    var location = path.pageInstances
        .where((i) => i is KeyPageInstance || i is AnonymousPage)
        .map(
      (i) {
        if (i is KeyPageInstance)
          return [i.keyPage.key, ...i.param].join("/");
        else
          return (i as AnonymousPage).keyNumber;
      },
    ).join("/");

    return RouteInformation(location: "/$location");
  }
}

class SplitRoutePath {
  final List<PageInstance> _pageInstances;

  List<PageInstance> get pageInstances => List.unmodifiable(_pageInstances);

  SplitRoutePath([List<PageInstance>? pageInstances])
      : _pageInstances = pageInstances ?? <PageInstance>[];

  SplitRoutePath._copy(this._pageInstances);

  @override
  String toString() {
    return pageInstances.map((i) => i.toString()).join(" -> ");
  }

  SplitRoutePath copy() {
    return SplitRoutePath._copy([]..addAll(pageInstances));
  }

  bool equals(SplitRoutePath path) {
    if (this.pageInstances.length != path.pageInstances.length) return false;
    for (var i = 0; i < this.pageInstances.length; i++) {
      if (this.pageInstances[i] != path.pageInstances[i]) {
        return false;
      }
    }
    return true;
  }
}
