import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/router/navigator_widgets/router_navigator_scaffold.dart';
import 'package:clean_android_design/src/router/requirement_builder.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';

class SplitRouterDelegate extends RouterDelegate<SplitRoutePath>
    with ChangeNotifier, PopNavigatorRouterDelegateMixin<SplitRoutePath> {
  static SplitRouterDelegate get instance => _instance!;
  static bool get hasInstance => _instance != null;

  static SplitRouterDelegate? _instance;

  SplitRouterDelegate.newInstance({
    required this.pageCollectionMap,
    void Function(PathTransformerCollection pathTransformerCollection)?
        modifyPathTransformerCollection,
    required this.sideMenuWidgetBuilder,
    required this.loadingWidgetBuilder,
    bool keepRoutePath = true,
  }) {
    if (!keepRoutePath || _instance == null) {
      routePath = SplitRoutePath([pageCollectionMap.pageHome.withoutParam()]);
    } else {
      initialRoutePath = instance.routePath.copy();
      routePath = instance.routePath.copy();
    }

    pathTransformerCollection =
        PathTransformerCollection(pageCollectionMap: pageCollectionMap);
    modifyPathTransformerCollection?.call(pathTransformerCollection);

    _instance = this;
  }

  final Widget Function(BuildContext context, SplitRoutePath path)
      sideMenuWidgetBuilder;
  final Widget Function(BuildContext context) loadingWidgetBuilder;

  late SplitRoutePath routePath;
  SplitRoutePath? initialRoutePath;

  GlobalKey<NavigatorState> get navigatorKey {
    var pos = routePath.pageInstances.last.position;
    if (pos == SplitRouterPosition.Dialog) {
      return navigatorKeyDialog;
    } else if (useOnlySingleNavigator) {
      return navigatorKeyLeft;
    } else {
      return pos == SplitRouterPosition.Left
          ? navigatorKeyLeft
          : navigatorKeyRight;
    }
  }

  final GlobalKey<NavigatorState> navigatorKeyLeft =
      GlobalKey<NavigatorState>();
  final GlobalKey<NavigatorState> navigatorKeyRight =
      GlobalKey<NavigatorState>();
  final GlobalKey<NavigatorState> navigatorKeyDialog =
      GlobalKey<NavigatorState>();
  final GlobalKey<RequirementBuilderState> requirementBuilderKey =
      GlobalKey<RequirementBuilderState>();

  final PageCollectionMap pageCollectionMap;
  late PathTransformerCollection pathTransformerCollection;

  SplitRoutePath get currentConfiguration {
    return routePath;
  }

  bool? _useOnlySingleNavigator;
  bool get useOnlySingleNavigator => _useOnlySingleNavigator ?? true;

  @override
  Widget build(BuildContext context) {
    _useOnlySingleNavigator =
        !kIsWeb || MediaQuery.of(context).size.longestSide < 1200;

    return RequirementBuilder(
      key: requirementBuilderKey,
      routePath: routePath,
      redirectOnFail: (invalidPath, pageInstance) {
        var oldPath = routePath.copy();
        routePath = pathTransformerCollection.transformAll(routePath);
        if (!routePath.equals(oldPath))
          Future.delayed(Duration.zero, () => notifyListeners());

        if (routePath.equals(invalidPath)) setNewPage(pageInstance);
      },
      builder: (context, routePath) {
        var oldPath = routePath.copy();
        routePath = pathTransformerCollection.transformAll(routePath);
        if (!routePath.equals(oldPath))
          Future.delayed(Duration.zero, () => notifyListeners());

        return RouterNavigatorScaffold(
          key: Key("RouterNavigatorScaffold"),
          navigatorKeyDialog: navigatorKeyDialog,
          navigatorKeyLeft: navigatorKeyLeft,
          navigatorKeyRight: navigatorKeyRight,
          routePath: routePath,
          notifyListeners: notifyListeners,
          pageCollectionMap: pageCollectionMap,
          sideMenuWidgetBuilder: sideMenuWidgetBuilder,
        );
      },
      loadingBuilder: loadingWidgetBuilder,
    );
  }

  bool get bothSidesOpen => routePath.pageInstances
      .any((i) => i.position == SplitRouterPosition.Right);

  @override
  Future<void> setInitialRoutePath(SplitRoutePath configuration) {
    return setNewRoutePath(initialRoutePath ?? configuration);
  }

  @override
  Future<void> setNewRoutePath(
    SplitRoutePath path, {
    bool ignoreLock = false,
  }) async {
    if (LockNavigatorService.instance.isLocked && !ignoreLock) {
      LockNavigatorService.instance.animate(showSnackbar: true);
    } else {
      routePath = path;

      routePath = pathTransformerCollection.transformAll(routePath);
    }
  }

  Future<void> setNewInternalUri(Uri uri) async {
    var path = SplitRouteInformationParser.instance.parseInternalUri(uri);
    await setNewRoutePath(path, ignoreLock: true);

    notifyListeners();
  }

  Future<void> setNewPage(PageInstance page) async {
    routePath = SplitRoutePath([page]);
    routePath = pathTransformerCollection.transformAll(routePath);

    notifyListeners();
  }

  Future<void> pushPage(PageInstance page) {
    routePath = SplitRoutePath([]
      ..addAll(routePath.pageInstances)
      ..add(page));
    routePath = pathTransformerCollection.transformAll(routePath);
    notifyListeners();

    return page.completer.future;
  }

  Future<void> pushDialog(Widget dialog) {
    return pushPage(AnonymousPage(() => dialog, SplitRouterPosition.Dialog));
  }

  void pop([SplitRouterPosition? position]) {
    var lambda = (x) => position == null ? true : (x.position == position);
    var _pageInstances = <PageInstance>[]..addAll(routePath.pageInstances);

    if (position == SplitRouterPosition.Left) {
      for (var i = (_pageInstances.length - 1); i >= 0; i--) {
        if (_pageInstances[i].position == SplitRouterPosition.Right) {
          var removed = _pageInstances.removeAt(i);
          if (!removed.completer.isCompleted) removed.completer.complete();
        } else if (_pageInstances[i].position == SplitRouterPosition.Left) {
          var removed = _pageInstances.removeAt(i);
          if (!removed.completer.isCompleted) removed.completer.complete();
          break;
        }
      }
    } else if (_pageInstances.any(lambda)) {
      var removeInstance = _pageInstances.lastWhere(lambda);
      if (!removeInstance.completer.isCompleted)
        removeInstance.completer.complete();
      _pageInstances.remove(removeInstance);
    }
    routePath = SplitRoutePath(_pageInstances);

    routePath = pathTransformerCollection.transformAll(routePath);
    notifyListeners();
  }

  void refresh() {
    requirementBuilderKey.currentState?.recheckRequirements();
  }
}

enum SplitRouterPosition {
  Right,
  Left,
  Draggable,
  Dialog,
}

class PageCache {
  static PageCache instance = PageCache._();

  PageCache._();

  final Map<PageInstance, Page> pages = <PageInstance, Page>{};

  void clean(List<PageInstance> currentInstances) {
    for (var key in pages.keys) {
      if (!currentInstances.any((x) => x == x)) {
        pages.remove(key);
      }
    }
  }

  Page getCachedOrCreate(PageInstance instance, Page Function() create) {
    if (!pages.containsKey(instance)) {
      pages[instance] = create();
    }
    return pages[instance]!;
  }
}
