import 'dart:async';
import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';
import 'package:scroll_to_index/scroll_to_index.dart';

class CustomScrollBar<T> extends StatefulWidget {
  final AutoScrollController scrollController;
  final Map<T, int> titleIndices;
  final Widget Function(BuildContext context, T item)
      customScrollbarSmallIndicatorBuilder;
  final Widget Function(BuildContext context, T item)
      customScrollbarBigIndicatorBuilder;

  const CustomScrollBar({
    Key? key,
    required this.scrollController,
    required this.titleIndices,
    required this.customScrollbarSmallIndicatorBuilder,
    required this.customScrollbarBigIndicatorBuilder,
  }) : super(key: key);

  @override
  _CustomScrollBarState<T> createState() => _CustomScrollBarState<T>();
}

class _CustomScrollBarState<T> extends State<CustomScrollBar<T>> {
  GlobalKey scrollBarColumnKey = GlobalKey();

  bool mouseIsInRegion = false;
  bool scrollBarVisible = false;
  Timer? scrollBarTimer;

  double? indicatorPosTop;
  int? selectedSectionIndex;

  @override
  void initState() {
    super.initState();

    widget.scrollController.addListener(() {
      if (!scrollBarVisible && mounted) setState(() => scrollBarVisible = true);
      scrollBarTimer?.cancel();

      if (!mouseIsInRegion)
        scrollBarTimer = Timer(
            Duration(milliseconds: 2000),
            () => {
                  if (scrollBarVisible && mounted)
                    setState(() => scrollBarVisible = false)
                });
    });
  }

  void _setIndicator(double y) {
    var box =
        scrollBarColumnKey.currentContext!.findRenderObject()! as RenderBox;
    var scrollBarHeight = box.size.height;

    var yInColumn = max(min(y, scrollBarHeight), 0.0);

    setState(() {
      indicatorPosTop =
          max(min(yInColumn - 22, scrollBarHeight - 44), 0.0) + 20 + 3;
      var itemSize = box.size.height / widget.titleIndices.length;
      var currentItem = max(0,
          min(widget.titleIndices.length - 1, (yInColumn / itemSize).floor()));

      if (selectedSectionIndex != currentItem) {
        widget.scrollController.scrollToIndex(
          widget.titleIndices.values.toList()[currentItem],
          duration: Duration(milliseconds: 100),
          preferPosition: AutoScrollPosition.begin,
        );
      }

      selectedSectionIndex = currentItem;
    });

    scrollBarTimer?.cancel();

    if (!mouseIsInRegion)
      scrollBarTimer = Timer(
          Duration(milliseconds: 2000),
          () =>
              {if (scrollBarVisible) setState(() => scrollBarVisible = false)});
  }

  @override
  Widget build(BuildContext context) {
    return Stack(
      children: [
        Align(
          alignment: Alignment.centerRight,
          child: Container(
            width: 50,
            child: MouseRegion(
              onEnter: (_) => setState(() {
                scrollBarTimer?.cancel();
                mouseIsInRegion = true;
                scrollBarVisible = true;
              }),
              onExit: (_) => setState(() {
                mouseIsInRegion = false;
                scrollBarTimer = Timer(
                    Duration(milliseconds: 2000),
                    () => {
                          if (scrollBarVisible)
                            setState(() => scrollBarVisible = false)
                        });
              }),
              opaque: false,
              child: Column(
                mainAxisSize: MainAxisSize.max,
              ),
            ),
          ),
        ),
        Align(
            alignment: Alignment.centerRight,
            child: AnimatedPadding(
              duration: Duration(milliseconds: 800),
              padding: EdgeInsets.only(
                  right: 10,
                  top: 20,
                  bottom: 15,
                  left: scrollBarVisible
                      ? 0
                      : (MediaQuery.of(context).size.width + 5)),
              child: GestureDetector(
                onVerticalDragUpdate: (details) =>
                    _setIndicator(details.localPosition.dy),
                onVerticalDragEnd: (details) =>
                    setState(() => indicatorPosTop = null),
                onTapDown: (details) {
                  selectedSectionIndex = null;
                  _setIndicator(details.localPosition.dy);
                },
                onTapUp: (details) => setState(() => indicatorPosTop = null),
                child: Container(
                  decoration: BoxDecoration(
                    color: DefaultColors.backgroundOutline.withOpacity(0.7),
                    borderRadius: BorderRadius.circular(10),
                  ),
                  width: 22,
                  padding: EdgeInsets.symmetric(vertical: 3),
                  child: DefaultTextStyle(
                    style: TextStyles.defaultTextStyle().copyWith(
                      fontSize: 12,
                      fontWeight: FontWeight.w400,
                    ),
                    child: Column(
                      key: scrollBarColumnKey,
                      mainAxisAlignment: MainAxisAlignment.spaceAround,
                      children: buildSmallIndicators(context),
                    ),
                  ),
                ),
              ),
            )),
        indicatorPosTop == null || !scrollBarVisible
            ? Container()
            : AnimatedPositioned(
                duration: Duration(milliseconds: 50),
                right: 11,
                top: indicatorPosTop,
                child: Container(
                  width: 20,
                  height: 44,
                  decoration: BoxDecoration(
                    color: DefaultColors.accent.withOpacity(0.6),
                    borderRadius: BorderRadius.circular(10),
                  ),
                ),
              ),
        indicatorPosTop == null || !scrollBarVisible
            ? Container()
            : AnimatedPositioned(
                duration: Duration(milliseconds: 50),
                right: 50,
                top: indicatorPosTop,
                child: Container(
                  height: 45,
                  decoration: BoxDecoration(
                    color: DefaultColors.accent.withOpacity(0.6),
                    borderRadius: BorderRadius.circular(50),
                  ),
                  padding: EdgeInsets.all(8.0),
                  child: Center(
                    child: DefaultTextStyle(
                      style: TextStyles.defaultTextStyle().copyWith(
                        color: DefaultColors.backgroundItem,
                        fontSize: 17,
                      ),
                      child: widget.customScrollbarBigIndicatorBuilder(
                          context,
                          widget.titleIndices.keys
                              .toList()[selectedSectionIndex!]),
                    ),
                  ),
                ),
              )
      ],
    );
  }

  List<Widget> buildSmallIndicators(BuildContext context) {
    return widget.titleIndices
        .map((item, index) => MapEntry(
            widget.customScrollbarSmallIndicatorBuilder(context, item), index))
        .keys
        .toList();
  }
}
