import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class ListAppBar<T extends IdItemInterface> extends StatefulWidget
    implements PreferredSizeWidget {
  final Widget? title;
  final Widget? selectionTitle;
  final bool showCountInDescription;
  final Widget? description;

  final bool? showBackArrow;
  final bool showSelectAllButton;
  final bool hasSearch;

  final List<Widget?>? appBarButtons;
  final List<Widget?>? appBarPopupButtons;

  final ListState? customListState;

  const ListAppBar({
    Key? key,
    this.title,
    this.showBackArrow,
    this.showSelectAllButton = false,
    this.appBarButtons,
    this.appBarPopupButtons,
    this.showCountInDescription = false,
    this.selectionTitle,
    this.description,
    this.hasSearch = true,
  })  : customListState = null,
        super(key: key);

  const ListAppBar.withCustomListState({
    Key? key,
    this.title,
    this.showBackArrow,
    this.showSelectAllButton = false,
    this.appBarButtons,
    this.appBarPopupButtons,
    this.showCountInDescription = false,
    this.selectionTitle,
    this.description,
    this.hasSearch = true,
    required this.customListState,
  }) : super(key: key);

  @override
  State<ListAppBar<T>> createState() => _ListAppBarState<T>();

  @override
  Size get preferredSize => AppBar(
        title: Text("Test"),
      ).preferredSize;
}

class _ListAppBarState<T extends IdItemInterface> extends State<ListAppBar<T>> {
  FocusNode searchFocusNode = FocusNode();

  void listener() {
    if (mounted) setState(() {});
  }

  ListState get listState =>
      widget.customListState ?? ListController.of(context).state;

  @override
  Widget build(BuildContext context) {
    listState.removeListener(listener);
    listState.addListener(listener);

    return AnimatedSwitcher(
      duration: Duration(milliseconds: 300),
      child: listState.searchIsActive && widget.hasSearch
          ? buildSearchAppBar(context)
          : (listState.selectionIsActive
              ? buildSelectionAppBar(context)
              : buildDefaultAppBar(context)),
    );
  }

  Widget buildSearchAppBar(BuildContext context) {
    Future.delayed(Duration.zero, () => searchFocusNode.requestFocus());
    return DefaultAppBar(
      key: ValueKey("search"),
      showBackArrow: false,
      titleFullSpace: true,
      title: SearchBar(
        controller: listState.searchController,
        focusNode: searchFocusNode,
      ),
    );
  }

  Widget buildSelectionAppBar(BuildContext context) {
    var length = listState.selectedRefs.length;
    return DefaultAppBar(
      key: ValueKey("selection"),
      title: widget.selectionTitle ?? Text("${length.toString()} ausgewählt"),
      showBackArrow: true,
      appBarButtons: [buildSearchButton(context), getSelectionAppbarButton()],
      appBarPopupButtons: getAppBarPopupButtons(context),
    );
  }

  Widget buildDefaultAppBar(BuildContext context) {
    return DefaultAppBar(
      key: ValueKey("default"),
      title: widget.title,
      description: widget.description ??
          (listState.allReferences != null && widget.showCountInDescription
              ? Text(listState.allReferences!.length.toString() + " angezeigt")
              : null),
      showBackArrow: widget.showBackArrow,
      appBarButtons: [
        buildSearchButton(context),
        ...(widget.appBarButtons ?? [])
      ],
      appBarPopupButtons: getAppBarPopupButtons(context),
    );
  }

  List<Widget?> getAppBarPopupButtons(BuildContext context) {
    var list = <Widget?>[];

    if (widget.appBarPopupButtons != null)
      list.addAll(widget.appBarPopupButtons!);

    if (widget.showSelectAllButton &&
        !listState.selectionIsActive &&
        listState.multiSelect) list.add(buildSelectAllButton(context));

    if (listState.selectionIsActive && widget.appBarButtons != null)
      list.addAll(widget.appBarButtons!);

    return list;
  }

  Widget buildSelectAllButton(BuildContext context) {
    return IconButton(
      icon: Icon(Icons.check_circle),
      onPressed: () {
        listState.selectAll();
      },
      tooltip: "Alles auswählen",
    );
  }

  Widget? buildSearchButton(BuildContext context) {
    return widget.hasSearch
        ? IconButton(
            icon: Icon(
              Icons.search,
            ),
            onPressed: () {
              listState.searchIsActive = true;
            },
            tooltip: "Suchen",
          )
        : null;
  }

  Widget? getSelectionAppbarButton() {
    if (!listState.multiSelect) return null;

    var all = listState.selectedRefs.length ==
        (listState.allReferences?.length ?? -1);
    var none = listState.selectedRefs.length == 0;

    return IconButton(
      tooltip: all ? "Auswahl verwerfen" : "Alle auswählen",
      onPressed: () => setState(() {
        if (all) {
          listState.selectedRefs.clear();
          listState.notifyListeners();
        } else {
          listState.selectAll();
        }
      }),
      icon: Icon(
        all
            ? Icons.check_circle
            : (none
                ? Icons.radio_button_unchecked
                : Icons.check_circle_outline),
        color: all ? DefaultColors.accent : DefaultColors.text,
      ),
    );
  }
}
