import 'dart:async';
import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';
import 'package:scroll_to_index/scroll_to_index.dart';

class ListContentWidget<T extends IdItemInterface> extends StatefulWidget {
  final List<T> Function(BuildContext context) listBuilder;
  final Widget Function(BuildContext context, T item) itemBuilder;
  final Widget Function(T)? groupTitle;

  final Widget? Function(BuildContext context)? topWidgetBuilder;
  final Widget? Function(BuildContext context)? bottomWidgetBuilder;
  final Widget? Function(BuildContext context)? noItemWidgetBuilder;

  final List<Comparable> Function(T)? sortProperties;
  final bool reverseSort;
  final Comparable Function(T)? groupProperty;
  final List<String> Function(T item)? searchArguments;

  final bool itemsAreSelectable;
  final IListableCollection collection;

  final bool isItemPicker;
  final bool multiSelectPicker;
  final void Function(List<DocumentReference> refs)? saveRefsToItem;

  final bool withCustomScrollbar;
  final Widget Function(BuildContext context, T item)?
      customScrollbarSmallIndicatorBuilder;
  final Widget Function(BuildContext context, T item)?
      customScrollbarBigIndicatorBuilder;

  final void Function(ScrollEdge edge)? scrollEdgeListener;

  const ListContentWidget({
    Key? key,
    required this.itemBuilder,
    this.groupTitle,
    this.topWidgetBuilder,
    this.bottomWidgetBuilder,
    this.noItemWidgetBuilder,
    this.sortProperties,
    required this.reverseSort,
    this.groupProperty,
    required this.itemsAreSelectable,
    required this.isItemPicker,
    required this.multiSelectPicker,
    this.saveRefsToItem,
    required this.listBuilder,
    this.searchArguments,
    required this.collection,
    this.withCustomScrollbar = false,
    this.customScrollbarSmallIndicatorBuilder,
    this.customScrollbarBigIndicatorBuilder,
    this.scrollEdgeListener,
  }) : super(key: key);

  @override
  State<ListContentWidget<T>> createState() => _ListContentWidgetState<T>();
}

class _ListContentWidgetState<T extends IdItemInterface>
    extends State<ListContentWidget<T>> {
  final scrollController = AutoScrollController();

  ScrollEdge? lastScrollEdge;

  @override
  void initState() {
    super.initState();

    if (widget.scrollEdgeListener != null)
      scrollController.addListener(() {
        checkScrollEdge();
      });
  }

  void checkScrollEdge({int recursionDepth = 0}) {
    if (widget.scrollEdgeListener == null) {
      return;
    }

    if (scrollController.positions.isEmpty) {
      if (recursionDepth < 2) {
        Future.delayed(Duration.zero,
            () => checkScrollEdge(recursionDepth: recursionDepth + 1));
      }
      return;
    }

    if (scrollController.position.pixels == 0 &&
        scrollController.position.maxScrollExtent == 0 &&
        scrollController.position.minScrollExtent == 0) {
      if (lastScrollEdge != ScrollEdge.BothEdges) {
        widget.scrollEdgeListener?.call(lastScrollEdge = ScrollEdge.BothEdges);
      }
    } else {
      if (scrollController.position.atEdge) {
        if (scrollController.position.pixels == 0) {
          if (lastScrollEdge != ScrollEdge.TopEdge) {
            widget.scrollEdgeListener
                ?.call(lastScrollEdge = ScrollEdge.TopEdge);
          }
        } else {
          if (lastScrollEdge != ScrollEdge.BottomEdge) {
            widget.scrollEdgeListener
                ?.call(lastScrollEdge = ScrollEdge.BottomEdge);
          }
        }
      } else {
        if (lastScrollEdge != ScrollEdge.InBetween) {
          widget.scrollEdgeListener
              ?.call(lastScrollEdge = ScrollEdge.InBetween);
        }
      }
    }
  }

  @override
  Widget build(BuildContext context) {
    ListController.of(context).state.removeListener(listener);
    ListController.of(context).state.addListener(listener);

    if (widget.scrollEdgeListener != null) {
      checkScrollEdge();
    }

    if (ListController.of(context).state.searchIsActive) {
      return buildForSearch();
    } else {
      return buildDefault();
    }
  }

  List<List<T>> getSortedList() {
    List<T> list = widget.listBuilder(context);
    if (widget.sortProperties != null) {
      list.sort((item1, item2) {
        var p1 = widget.sortProperties!(widget.reverseSort ? item2 : item1);
        var p2 = widget.sortProperties!(widget.reverseSort ? item1 : item2);
        for (var i = 0; i < p1.length; i++) {
          var result = p1[i].compareTo(p2[i]);
          if (result != 0) return result;
        }
        return 0;
      });
    }

    var groupedItems = <List<T>>[];
    if (widget.groupProperty == null) {
      groupedItems.add(list);
    } else if (list.length > 0) {
      var lastList = <T>[];
      groupedItems.add(lastList);
      list.asMap().forEach((i, item) {
        if (i == 0)
          lastList.add(item);
        else if (widget.groupProperty!(lastList.last)
                .compareTo(widget.groupProperty!(item)) ==
            0)
          lastList.add(item);
        else {
          lastList = <T>[];
          lastList.add(item);
          groupedItems.add(lastList);
        }
      });
    }

    return groupedItems;
  }

  List<T?> flattenSortedList(List<List<T>> sortedList) {
    return (<List<T>>[[]]..addAll(sortedList))
        .cast<List<T?>>()
        .reduce((value, element) {
      return <T?>[]
        ..addAll(value)
        ..add(null)
        ..addAll(element);
    });
  }

  Map<T, int> getTitleIndices(List<T?> flattendList) {
    Map<T, int> result = Map<T, int>();
    flattendList.asMap().forEach((index, item) {
      if (item == null &&
          index < flattendList.length - 1 &&
          flattendList[index + 1] != null) {
        result[flattendList[index + 1]!] = index;
      }
    });
    return result;
  }

  Widget buildDefault() {
    var sortedList = getSortedList();
    var list = flattenSortedList(sortedList);

    Map<T, int>? titleIndices =
        widget.withCustomScrollbar ? getTitleIndices(list) : null;

    ListController.of(context).state.allReferences = list
        .whereNotNull()
        .map((item) => widget.collection.ref(item.id!))
        .toList();

    return ScrollConfiguration(
      behavior: ScrollConfiguration.of(context)
          .copyWith(scrollbars: !widget.withCustomScrollbar),
      child: Stack(
        children: [
          CustomScrollView(
            controller: scrollController,
            shrinkWrap: true,
            slivers: [
              widget.topWidgetBuilder != null
                  ? SliverPadding(
                      padding: EdgeInsets.only(bottom: 10),
                      sliver: SliverToBoxAdapter(
                        child: widget.topWidgetBuilder!.call(context),
                      ),
                    )
                  : null,
              list.whereNotNull().length > 0
                  ? SliverList(
                      delegate: SliverChildBuilderDelegate(
                        (context, index) {
                          Widget getChild() {
                            if (list[index] == null) {
                              if (index < list.length - 1 &&
                                  list[index + 1] != null) {
                                return ItemSectionTitle(
                                  title: widget.groupTitle
                                          ?.call(list[index + 1]!) ??
                                      Container(),
                                );
                              } else {
                                return Container();
                              }
                            } else {
                              return buildItem(context, list, index);
                            }
                          }

                          return AutoScrollTag(
                            key: ValueKey(index),
                            controller: scrollController,
                            index: index,
                            child: getChild(),
                          );
                        },
                        childCount: list.length,
                      ),
                    )
                  : SliverToBoxAdapter(
                      child: DescriptionText.fromChild(
                          widget.noItemWidgetBuilder != null
                              ? (widget.noItemWidgetBuilder?.call(context) ??
                                  Container())
                              : Text("Keine Einträge verfügbar")),
                    ),
              widget.bottomWidgetBuilder != null
                  ? SliverPadding(
                      padding: EdgeInsets.only(top: 5),
                      sliver: SliverToBoxAdapter(
                        child: widget.bottomWidgetBuilder!.call(context),
                      ),
                    )
                  : null,
            ].whereNotNull(),
          ),
          buildScrollBar(titleIndices),
        ],
      ),
    );
  }

  Widget buildScrollBar(Map<T, int>? titleIndices) {
    if (!widget.withCustomScrollbar) return Container();

    assert(
        widget.customScrollbarSmallIndicatorBuilder != null &&
            widget.customScrollbarBigIndicatorBuilder != null,
        "[ListContentWidget] Used CustomScollBar without providing the indicator builder!");

    return CustomScrollBar<T>(
      scrollController: scrollController,
      titleIndices: titleIndices!,
      customScrollbarSmallIndicatorBuilder:
          widget.customScrollbarSmallIndicatorBuilder!,
      customScrollbarBigIndicatorBuilder:
          widget.customScrollbarBigIndicatorBuilder!,
    );
  }

  Widget buildItem(BuildContext context, List<T?> list, int index) {
    var item = list[index]!;
    return ItemListBehavior(
      isSelected: ListController.of(context).state.isSelected(item),
      onTapList: () {
        var state = ListController.of(context).state;
        if (state.selectionIsActive) {
          state.toogleSelectionFor(widget.collection.ref(item.id!));
          if (widget.isItemPicker && !widget.multiSelectPicker) {
            widget.saveRefsToItem?.call([widget.collection.ref(item.id!)]);
            Navigator.of(context).pop();
          }
          return false;
        } else {
          return true;
        }
      },
      onLongTapList: () {
        var state = ListController.of(context).state;
        if (widget.itemsAreSelectable || state.selectionIsActive) {
          state.toogleSelectionFor(widget.collection.ref(item.id!));
          return false;
        } else {
          return true;
        }
      },
      child: ItemPosition(
        child: widget.itemBuilder(context, item),
        isFirst: index == 0 || list[index - 1] == null,
        isLast: index == (list.length - 1) || list[index + 1] == null,
        hasSeparationLine:
            (index == (list.length - 1) || list[index + 1] == null),
      ),
    );
  }

  List<T> getSearchList() {
    var searchTexts = ListController.of(context)
        .state
        .searchController
        .text
        .toLowerCase()
        .split(" ")
      ..removeWhere((s) => s == "");

    double _calcSearchResults(T item) {
      if (searchTexts.length == 0 || widget.searchArguments == null) return 1;
      var scores = List.generate(searchTexts.length, (index) => 0.0);
      searchTexts.asMap().forEach((n, searchText) {
        var args = widget.searchArguments!(item)
            .map((t) => (t).toLowerCase())
            .toList();
        for (var i = 0; i < args.length; i++) {
          if (args[i].contains(searchText)) {
            if (args[i].startsWith(searchText)) {
              scores[n] = max(scores[n], (args.length - i) * 10 + 1.0);
            } else {
              scores[n] = max(scores[n], (args.length - i) * 10 + 0.5);
            }
          }
        }
      });
      return scores.reduce(min) == 0 ? 0 : scores.reduce(max);
    }

    List<T> list = widget.listBuilder(context);

    var searchResults = Map<T, double>();
    if (searchTexts.length > 0) {
      searchResults.updateAll((key, value) => 0);
      list.forEach((item) => searchResults.update(
            item,
            (_) => _calcSearchResults(item),
            ifAbsent: () => _calcSearchResults(item),
          ));
    } else {
      searchResults.updateAll((key, value) => 0);
    }

    return list.where((item) => (searchResults[item] ?? 0) > 0).toList()
      ..sort((item1, item2) {
        if (searchResults[item2]!.compareTo(searchResults[item1]!) != 0)
          return searchResults[item2]!.compareTo(searchResults[item1]!);

        if (widget.sortProperties != null) {
          var p1 = widget.sortProperties!(widget.reverseSort ? item2 : item1);
          var p2 = widget.sortProperties!(widget.reverseSort ? item1 : item2);
          for (var i = 0; i < p1.length; i++) {
            var result = p1[i].compareTo(p2[i]);
            if (result != 0) return result;
          }
        }
        return 0;
      });
  }

  Widget buildForSearch() {
    return ValueListenableBuilder<TextEditingValue>(
        valueListenable: ListController.of(context).state.searchController,
        builder: (context, value, _) {
          var list = getSearchList();
          ListController.of(context).state.allReferences =
              list.map((item) => widget.collection.ref(item.id!)).toList();

          return CustomScrollView(
            shrinkWrap: true,
            slivers: [
              widget.topWidgetBuilder != null
                  ? SliverPadding(
                      padding: EdgeInsets.only(bottom: 10),
                      sliver: SliverToBoxAdapter(
                        child: widget.topWidgetBuilder!.call(context),
                      ),
                    )
                  : null,
              list.length > 0
                  ? SliverList(
                      delegate: SliverChildBuilderDelegate(
                        (context, index) {
                          return buildItem(context, list, index);
                        },
                        childCount: list.length,
                      ),
                    )
                  : SliverToBoxAdapter(
                      child: DescriptionText(ListController.of(context)
                                  .state
                                  .searchController
                                  .text
                                  .trim() !=
                              ""
                          ? "Kein Suchergebnis gefunden"
                          : "Bitte einen Suchtext eingeben"),
                    ),
            ].whereNotNull(),
          );
        });
  }

  void listener() {
    setState(() {});
  }
}

enum ScrollEdge {
  TopEdge,
  BottomEdge,
  BothEdges,
  InBetween,
}
