import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class DefaultAppBar extends StatelessWidget implements PreferredSizeWidget {
  final Widget? title;
  final Widget? description;

  final bool? showBackArrow;

  final List<Widget?>? appBarButtons;
  final List<Widget?>? appBarPopupButtons;

  final bool titleFullSpace;

  const DefaultAppBar({
    Key? key,
    this.showBackArrow,
    this.appBarButtons,
    this.appBarPopupButtons,
    this.titleFullSpace = false,
    required this.title,
    this.description,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return AppBar(
      leading: buildLeading(context),
      actions: getAppBarActions(context),
      backgroundColor: DefaultColors.background,
      elevation: 0,
      automaticallyImplyLeading: false,
      actionsIconTheme: IconThemeData(
        size: 23,
        color: DefaultColors.text,
      ),
      titleSpacing: titleFullSpace ? 0 : NavigationToolbar.kMiddleSpacing,
      iconTheme: IconThemeData(
        color: DefaultColors.text,
      ),
      title: getTitle(),
    );
  }

  Widget getTitle() {
    return Column(
      mainAxisAlignment: MainAxisAlignment.center,
      crossAxisAlignment: CrossAxisAlignment.start,
      mainAxisSize: MainAxisSize.min,
      children: [
        DefaultTextStyle(
          style:
              TextStyles.textStyleSize(19).copyWith(color: DefaultColors.text),
          child: title ?? Container(),
          overflow: TextOverflow.ellipsis,
        ),
        DefaultTextStyle(
          style: TextStyles.textStyleSize(13)
              .copyWith(color: DefaultColors.textLight),
          child: description ?? Container(),
          overflow: TextOverflow.ellipsis,
        ),
      ],
    );
  }

  Widget? buildLeading(BuildContext context) {
    var list = <Widget>[];
    if (showBackArrow ?? Navigator.of(context).canPop()) {
      list.add(IconButton(
        icon: Icon(
          Icons.keyboard_arrow_left,
          color: DefaultColors.text,
          size: 30,
        ),
        onPressed: () async {
          Navigator.of(context).maybePop();
        },
      ));
    }

    if (Scaffold.of(context).hasDrawer) {
      list.add(IconButton(
        icon: Icon(
          Icons.menu,
          color: DefaultColors.text,
          size: 24,
        ),
        tooltip: "Navigation öffnen",
        onPressed: () async {
          Scaffold.of(context).openDrawer();
        },
      ));
    }

    return list.length > 0
        ? ExcludeFocusTraversal(
            child: Row(
              children: list,
              mainAxisSize: MainAxisSize.min,
              mainAxisAlignment: MainAxisAlignment.start,
            ),
          )
        : null;
  }

  List<Widget>? getAppBarActions(BuildContext context) {
    var list = (appBarButtons?.reversed.toList() ?? <Widget>[])
        .where((x) => (!(x is Container) || x.child != null))
        .toList()
        .whereNotNull();

    double titleWidth = MediaQuery.of(context).size.width;

    if (title is Text)
      titleWidth = (title as Text).getLayoutWidth(
        context,
        explicitTextStyle: TextStyles.textStyleSize(19),
      );
    if (title is SizedBox) titleWidth = (title as SizedBox).width ?? 50;

    if (description is Text)
      titleWidth = max(
          titleWidth,
          (description as Text).getLayoutWidth(
            context,
            explicitTextStyle: TextStyles.textStyleSize(13),
          ));
    if (description is SizedBox)
      titleWidth = max(titleWidth, (description as SizedBox).width ?? 50);

    var screenWidth = MediaQuery.of(context).size.width;
    var spaceForWidgets = max(
        0,
        screenWidth -
            titleWidth -
            45.0 -
            ((showBackArrow ?? Navigator.of(context).canPop()) ? 48 : 0));
    var maxWidgets = min(5, (spaceForWidgets / 48.0).floor());

    var newList = <Widget>[];
    var forMenu = <Widget>[];

    forMenu.addAll(appBarPopupButtons
            ?.where((x) => (!(x is Container) || x.child != null))
            .toList()
            .whereNotNull() ??
        <Widget>[]);

    if (forMenu.length > 0) maxWidgets--;

    if (list.length <= maxWidgets || list.length == 0) {
      newList.addAll(list);
    } else {
      newList
          .addAll(list.sublist(min(list.length, list.length - maxWidgets + 1)));
      forMenu.addAll(
          list.sublist(0, min(list.length, list.length - maxWidgets + 1)));
    }

    newList = newList.reversed.toList();

    if (forMenu.length > 0)
      newList.add(PopupMenuButton<void Function()>(
        icon: Icon(Icons.more_vert),
        onSelected: (value) => value.call(),
        tooltip: "Mehr zeigen",
        itemBuilder: (context) => forMenu
            .map(
              (item) {
                assert(
                    item is IconButton ||
                        item is PopupMenuEntry<void Function()> ||
                        item is PopupMenuButton<void Function()> ||
                        item is AppBarAction,
                    "Only IconButtons can be in the app bar when wrapping in a pop-up menu");

                if (item is PopupMenuEntry<void Function()>) {
                  return item;
                }

                if (item is IconButton) {
                  return PopupMenuItem(
                    value: item.onPressed,
                    child: IconTheme(
                      data: IconThemeData(size: 23, color: DefaultColors.text),
                      child: Row(
                        children: [
                          item.icon,
                          Container(width: 8),
                          Text(
                            item.tooltip!,
                            style: TextStyle(color: DefaultColors.text),
                          )
                        ],
                      ),
                    ),
                  );
                }

                if (item is AppBarAction) {
                  return PopupMenuItem<void Function()>(
                    value: (item as AppBarAction).onPressed,
                    child: IconTheme(
                      data: IconThemeData(size: 23, color: DefaultColors.text),
                      child: Row(
                        children: [
                          (item as AppBarAction).icon,
                          Container(width: 8),
                          Text(
                            (item as AppBarAction).tooltip,
                            style: TextStyle(color: DefaultColors.text),
                          )
                        ],
                      ),
                    ),
                  );
                }

                if (item is PopupMenuButton<void Function()>) {
                  return PopupMenuItem(
                    value: () {},
                    child: PopupMenuButton<void Function()>(
                      tooltip: item.tooltip,
                      child: Padding(
                        child: IconTheme(
                          data: IconThemeData(size: 23),
                          child: Row(
                            children: [
                              item.icon!,
                              Container(width: 8),
                              Text(item.tooltip!)
                            ],
                          ),
                        ),
                        padding: EdgeInsets.symmetric(vertical: 8.0),
                      ),
                      onSelected: item.onSelected,
                      itemBuilder: item.itemBuilder,
                    ),
                  );
                }

                return PopupMenuItem<void Function()>(
                  child: null,
                );
              },
            )
            .toList()
            .reversed
            .toList(),
      ));
    return newList;
  }

  @override
  Size get preferredSize => AppBar(
        title: Text("Test"),
      ).preferredSize;
}

abstract class AppBarAction {
  void Function()? get onPressed;
  Icon get icon;
  String get tooltip;
}
