import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class HintWidget extends StatefulWidget {
  final Widget Function(Animation<double> animation) builder;
  final Future<void> Function()? onTap;
  final String? heroTag;
  final int maxWidth;
  final bool startOpen;

  final Color? backgroundColor;
  final Color? color;
  final Color? borderColor;

  const HintWidget({
    Key? key,
    required this.builder,
    this.onTap,
    this.heroTag,
    this.maxWidth = 85,
    this.startOpen = false,
    this.backgroundColor,
    this.borderColor,
    this.color,
  }) : super(key: key);

  @override
  _HintWidgetState createState() => _HintWidgetState();
}

class _HintWidgetState extends State<HintWidget>
    with SingleTickerProviderStateMixin {
  late AnimationController _animationController;
  late Animation<int> _widthTween;

  bool _waitForOpenNavigator = false;
  int? _waitForAnimationId;

  @override
  void initState() {
    super.initState();

    _animationController = AnimationController(
      vsync: this,
      duration: Duration(milliseconds: 300),
      reverseDuration: Duration(milliseconds: 100),
    );

    var curvedAnimation =
        CurvedAnimation(parent: _animationController, curve: Curves.ease);

    _widthTween =
        IntTween(begin: widget.maxWidth, end: 30).animate(curvedAnimation);

    _animationController.addStatusListener((status) {
      if (status == AnimationStatus.dismissed)
        _closeInDuration(Duration(milliseconds: 3000));
    });
  }

  _closeInDuration(Duration duration) {
    var id = Random().nextInt(100000);
    _waitForAnimationId = id;
    Future.delayed(duration, () {
      if (_animationController.status == AnimationStatus.dismissed &&
          !_waitForOpenNavigator &&
          _waitForAnimationId == id &&
          mounted) {
        _animationController.forward();
      }
    });
  }

  @override
  Widget build(BuildContext context) {
    if (widget.startOpen) {
      _closeInDuration(Duration(milliseconds: 2000));
    } else {
      _animationController.animateTo(1, duration: Duration(milliseconds: 0));
    }

    return Hero(
      tag: widget.heroTag ?? Random().nextInt(10000).toString(),
      child: ExcludeFocusTraversal(
        child: Material(
          color: widget.backgroundColor ?? DefaultColors.accent,
          borderRadius: BorderRadius.only(
              topLeft: Radius.circular(8), bottomLeft: Radius.circular(8)),
          child: MouseRegion(
            onEnter: (event) {
              if (_animationController.status == AnimationStatus.completed)
                _animationController.reverse();
            },
            child: InkWell(
              onTap: () {
                if (_animationController.status == AnimationStatus.completed)
                  _animationController.reverse();
                else
                  _showWidget();
              },
              borderRadius: BorderRadius.only(
                  topLeft: Radius.circular(8), bottomLeft: Radius.circular(8)),
              highlightColor: Color.lerp(
                  (widget.backgroundColor ?? DefaultColors.accent),
                  (widget.color ?? DefaultColors.backgroundItem),
                  0.2),
              child: AnimatedBuilder(
                animation: _widthTween,
                child: Container(),
                builder: (context, _) => Container(
                    height: 45,
                    width: _widthTween.value + 0.0,
                    decoration: BoxDecoration(
                      borderRadius: BorderRadius.only(
                          topLeft: Radius.circular(8),
                          bottomLeft: Radius.circular(8)),
                      border: Border.all(
                          color: widget.borderColor ?? DefaultColors.textLight),
                    ),
                    child: DefaultTextStyle(
                      style: TextStyle(
                        color: widget.color ?? DefaultColors.backgroundItem,
                      ),
                      child: IconTheme(
                        data: IconThemeData(
                          color: widget.color ?? DefaultColors.backgroundItem,
                        ),
                        child: widget.builder(_animationController),
                      ),
                    )),
              ),
            ),
          ),
        ),
      ),
    );
  }

  _showWidget() async {
    _waitForOpenNavigator = true;
    _animationController.reverse();

    await widget.onTap?.call();

    _waitForOpenNavigator = false;
    _closeInDuration(Duration(milliseconds: 400));

    setState(() {});
  }

  @override
  void dispose() {
    _animationController.dispose();
    super.dispose();
  }
}
