import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class ScaffoldPages extends StatefulWidget {
  final List<ScaffoldPage?> pages;

  const ScaffoldPages({Key? key, required this.pages}) : super(key: key);

  @override
  State<ScaffoldPages> createState() => ScaffoldPagesState();
}

class ScaffoldPagesState extends State<ScaffoldPages> {
  Key get currentPageKey {
    var key = ScaffoldPageController.of(context).currentPageKey;
    if (key == null || !widget.pages.any((x) => x?.key == key)) {
      assert(widget.pages.whereNotNull().length > 0,
          "[ScaffoldPages] ScaffoldPages needs to be provided with at least one page");
      ScaffoldPageController.of(context)
          .setPage(widget.pages.whereNotNull().first.key);
      return currentPageKey;
    } else {
      return key;
    }
  }

  void listener() {
    setState(() {});
  }

  @override
  Widget build(BuildContext context) {
    assert(ScaffoldPageController.isIn(context),
        "[ScaffoldPages] ScaffoldPages needs to be places inside of ScaffoldPageController");
    ScaffoldPageController.of(context).updateNotifier.removeListener(listener);
    ScaffoldPageController.of(context).updateNotifier.addListener(listener);

    return widget.pages.length > 1
        ? NestedScrollView(
            headerSliverBuilder: (context, _) => [buildSliverIconBar()],
            body: _buildMainContent(),
          )
        : _buildMainContent();
  }

  Widget buildSliverIconBar() {
    return widget.pages.length > 1
        ? SliverPadding(
            padding: EdgeInsets.only(top: 5, bottom: 15),
            sliver: SliverToBoxAdapter(
              child: Row(
                mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                children: widget.pages
                    .whereNotNull()
                    .map(
                      (p) => ScaffoldPageIconInformation(
                        isSelected: p.key == currentPageKey,
                        onClick: () {
                          ScaffoldPageController.of(context).setPage(p.key);
                          p.onShowPage?.call();
                        },
                        child: p.pageIconBuilder(context),
                      ),
                    )
                    .toList(),
              ),
            ),
          )
        : SliverToBoxAdapter(child: Container());
  }

  bool _isPreparating(ScaffoldPage currentPage) {
    var future = currentPage.preparationFuture?.call();

    if (future == null) {
      return false;
    } else {
      future.whenComplete(() {
        if (mounted && currentPageKey == currentPage.key) setState(() {});
      });
      return true;
    }
  }

  Widget _buildMainContent() {
    var currentPage = widget.pages.firstWhere((x) => x?.key == currentPageKey)!;
    Widget child;
    if (_isPreparating(currentPage)) {
      child =
          currentPage.builderPreparation?.call(context) ?? ScaffoldShimmer();
    } else {
      child = currentPage.builder.call(context);
    }

    if (currentPage.scrollable || widget.pages.length == 1) {
      return SingleChildScrollView(
        child: child,
      );
    } else {
      return child;
    }
  }
}

class ScaffoldPageIconInformation extends InheritedWidget {
  final ScaffoldPageIcon child;
  final void Function() onClick;
  final bool isSelected;

  ScaffoldPageIconInformation({
    required this.child,
    required this.onClick,
    required this.isSelected,
  }) : super(child: child);

  static ScaffoldPageIconInformation of(BuildContext context) {
    final ScaffoldPageIconInformation? result = context
        .dependOnInheritedWidgetOfExactType<ScaffoldPageIconInformation>();
    assert(result != null, 'No ScaffoldPageIconInformation found in context');
    return result!;
  }

  @override
  bool updateShouldNotify(covariant InheritedWidget oldWidget) => true;
}

class ScaffoldPage {
  final Key key;
  final VoidCallback? onShowPage;

  final ScaffoldPageIcon Function(BuildContext context) pageIconBuilder;

  final Widget Function(BuildContext context) builder;
  final Widget Function(BuildContext context)? builderPreparation;

  final Future<int>? Function()? preparationFuture;

  final bool scrollable;

  const ScaffoldPage({
    required this.key,
    required this.pageIconBuilder,
    required this.builder,
    this.onShowPage,
    this.preparationFuture,
    this.builderPreparation,
    this.scrollable = true,
  });
}

class ScaffoldPageIcon extends StatelessWidget {
  final Widget? icon;
  final Widget? title;
  final VoidCallback? onTap;
  final bool? isSelected;

  final bool connectToScaffoldPages;

  const ScaffoldPageIcon({
    Key? key,
    this.icon,
    this.title,
    this.onTap,
    this.isSelected = false,
  })  : connectToScaffoldPages = true,
        super(key: key);

  const ScaffoldPageIcon.withCustomTap({
    Key? key,
    this.icon,
    this.title,
    this.onTap,
    this.isSelected = false,
  })  : connectToScaffoldPages = false,
        super(key: key);

  @override
  Widget build(BuildContext context) {
    ScaffoldPageIconInformation? scaffoldPagesInformation =
        connectToScaffoldPages ? ScaffoldPageIconInformation.of(context) : null;

    var isSelected = (scaffoldPagesInformation == null
            ? this.isSelected
            : scaffoldPagesInformation.isSelected) ??
        false;

    var onTap = scaffoldPagesInformation == null
        ? this.onTap
        : scaffoldPagesInformation.onClick;

    double width = 0.0;

    if (title is Text) {
      final TextPainter textPainter = TextPainter(
          text: TextSpan(
              text: (title! as Text).data,
              style: TextStyles.textStyleSize(13).copyWith(
                  color: DefaultColors.text, fontWeight: FontWeight.w600)),
          maxLines: 1,
          textDirection: TextDirection.ltr)
        ..layout(minWidth: 0, maxWidth: double.infinity);
      width = textPainter.size.width;
    } else if (title is SizedBox) {
      width = (title as SizedBox).width ?? width;
    }

    return ExcludeFocusTraversal(
      child: Container(
        width: max(width, 50) + 10.0,
        child: Column(
          children: <Widget>[
            Container(
              decoration: BoxDecoration(
                boxShadow: isSelected
                    ? [
                        BoxShadow(
                          color: DefaultColors.isDarkModeOn
                              ? DefaultColors.accent.withOpacity(0.1)
                              : DefaultColors.text.withOpacity(0.3),
                          spreadRadius: 0.5,
                          blurRadius: 3,
                          offset: Offset(0, 1),
                        ),
                      ]
                    : [],
                borderRadius: BorderRadius.circular(50),
              ),
              child: Material(
                color: isSelected
                    ? (DefaultColors.isDarkModeOn
                        ? DefaultColors.accent.withOpacity(0.8)
                        : Color.lerp(DefaultColors.backgroundOutline,
                            DefaultColors.backgroundElevated, 0.3))
                    : DefaultColors.backgroundOutline,
                borderRadius: BorderRadius.circular(50),
                child: InkWell(
                  borderRadius: BorderRadius.circular(50),
                  onTap: onTap,
                  child: Container(
                    width: 50,
                    height: 50,
                    decoration: BoxDecoration(
                      borderRadius: BorderRadius.circular(50),
                    ),
                    child: IconTheme(
                      data: IconThemeData(
                          color: DefaultColors.isDarkModeOn && isSelected
                              ? DefaultColors.background
                              : DefaultColors.text),
                      child: icon ?? Container(),
                    ),
                  ),
                ),
              ),
            ),
            SizedBox(
              height: 7,
            ),
            GestureDetector(
              onTap: () {
                onTap?.call();
              },
              child: DefaultTextStyle(
                  style: TextStyles.textStyleSize(13).copyWith(
                      color: DefaultColors.text,
                      fontWeight:
                          isSelected ? FontWeight.w600 : FontWeight.w400),
                  child: title ?? Container()),
            )
          ],
        ),
      ),
    );
  }
}

class ScaffoldPageController extends InheritedWidget {
  final ForceUpdateKey updateNotifier;
  final ScaffoldPageControllerContent _content;

  Key? get currentPageKey => _content._currentPageKey;

  void setPage(Key key) {
    if (key != currentPageKey) {
      _content._currentPageKey = key;
      updateNotifier.update();
    }
  }

  ScaffoldPageController({
    required Widget child,
    ScaffoldPageControllerContent? content,
  })  : updateNotifier = ForceUpdateKey(),
        this._content = content ?? ScaffoldPageControllerContent(),
        super(child: child);

  @override
  bool updateShouldNotify(InheritedWidget oldWidget) => true;

  static ScaffoldPageController of(BuildContext context) {
    final ScaffoldPageController? result =
        context.dependOnInheritedWidgetOfExactType<ScaffoldPageController>();
    assert(result != null, 'No ScaffoldPageController found in context');
    return result!;
  }

  static bool isIn(BuildContext context) {
    return context
            .dependOnInheritedWidgetOfExactType<ScaffoldPageController>() !=
        null;
  }
}

class ScaffoldPageControllerContent {
  Key? _currentPageKey;
}
