import 'dart:math';

import 'package:flutter/material.dart';

extension TextWidth on Text {
  double getLayoutWidth(BuildContext context, {TextStyle? explicitTextStyle}) {
    assert(this.style != null || explicitTextStyle != null,
        "A TextStyle needs to be provided");

    if (this.data == null || this.data == "") return 0;

    var textWidget = Text.rich(
        TextSpan(text: this.data, style: explicitTextStyle ?? this.style));
    var textWidgetBuild = textWidget.build(context) as RichText;
    var renderObject = textWidgetBuild.createRenderObject(context)
      ..layout(BoxConstraints());
    var boxes = renderObject.getBoxesForSelection(
        TextSelection(baseOffset: 0, extentOffset: this.data!.length));
    var textWidth = boxes.last.right;

    return textWidth;
  }
}

extension NotNullList<T> on List<T?> {
  List<T> whereNotNull() {
    return whereType<T>().toList();
  }
}

extension SimilarityString on String? {
  double similarityTo(String? other) {
    var levenshtein = levenshteinDistanceTo(other);
    var a = (this ?? "").toUpperCase();
    var b = (other ?? "").toUpperCase();
    return 1 - levenshtein / (max(a.length, b.length));
  }

  int levenshteinDistanceTo(String? other) {
    var a = (this ?? "").toUpperCase();
    var b = (other ?? "").toUpperCase();
    int sa = a.length;
    int sb = b.length;
    int i, j, cost, min1, min2, min3;
    List<List<int>> d =
        List.generate(sa + 1, (int i) => List.generate(sb + 1, (_) => 0));
    if (a.length == 0) {
      return b.length;
    }
    if (b.length == 0) {
      return a.length;
    }
    for (i = 0; i <= sa; i++) d[i][0] = i;
    for (j = 0; j <= sb; j++) d[0][j] = j;
    for (i = 1; i <= a.length; i++)
      for (j = 1; j <= b.length; j++) {
        if (a[i - 1] == b[j - 1])
          cost = 0;
        else
          cost = 1;
        min1 = (d[i - 1][j] + 1);
        min2 = (d[i][j - 1] + 1);
        min3 = (d[i - 1][j - 1] + cost);
        d[i][j] = min(min1, min(min2, min3));
      }
    return d[a.length][b.length];
  }
}

extension SameAsDateTime on DateTime {
  bool isAtSameDayAs(DateTime other) {
    return this.year == other.year &&
        this.month == other.month &&
        this.day == other.day;
  }

  int get weekOfYear {
    var from = DateTime.utc(this.year, 1, 1);
    from = from.subtract(Duration(days: from.weekday - 1));
    var to = DateTime.utc(this.year, this.month, this.day);
    to = to.subtract(Duration(days: to.weekday - 1));
    return (to.difference(from).inDays / 7).ceil();
  }

  int get weekSinceEpoch {
    var from = DateTime.utc(2000, 1, 1);
    from = from.subtract(Duration(days: from.weekday - 1));
    var to = DateTime.utc(this.year, this.month, this.day);
    to = to.subtract(Duration(days: to.weekday - 1));
    return (to.difference(from).inDays / 7).ceil();
  }
}
